"""
Copyright 2023-present fretgfr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
import io
import mimetypes
import os
from dataclasses import dataclass
from typing import Any, Dict, List, Literal, Optional, Tuple, Type, Union

from .enums import OAuthProviderType, QuotaType, UserRole
from .http import HTTPClient, Route
from .utils import MISSING, build_avatar_payload, generate_quota_payload, guess_mimetype_by_magicnumber, parse_iso_timestamp

__all__ = (
    "File",
    "Folder",
    "User",
    "InviteUser",
    "Invite",
    "TagFile",
    "Tag",
    "URL",
    "UploadFile",
    "UploadResponse",
    "FileData",
    "PartialQuota",
    "UserQuota",
    "OAuthProvider",
    "Thumbnail",
    "UserViewSettings",
    "ServerVersionInfo",
    "UserStats",
)

JSON = Union[Dict[str, Any], List[Any], int, str, float, bool, Type[None]]


@dataclass
class File:
    """
    Represents a file stored on Zipline.

    Attributes
    ----------
    id: :class:`str`
        The internal id of the file in Zipline.
    created_at: :class:`datetime.datetime`
        When the file was created.
    updated_at: :class:`datetime.datetime`
        When the file was last updated.
    deletes_at: Optional[:class:`datetime.datetime`]
        The scheduled deletion time of the file, if applicable.
    favorite: :class:`bool`
        Whether the file is favorited.
    original_name: Optional[:class:`str`]
        The original name of the file, if stored.
    name: :class:`str`
        The name of the file.
    size: :class:`int`
        The size of the file in bytes.
    type: :class:`str`
        The MIME type of the file.
    views: :class:`int`
        The number of times the file has been viewed.
    max_views: Optional[:class:`int`]
        The maximum number of times the file can be viewed before being removed, if applicable.
    password: Optional[Union[:class:`str`, :class:`bool`]]
        Whether the file is password protected.
    folder_id: Optional[:class:`str`]
        The id of the :class:`~zipline.models.Folder` that this file resides in, if applicable.
    thumbnail: Optional[:class:`~zipline.models.Thumbnail`]
        The thumbnail of the file, if available.
    tags: List[:class:`~zipline.models.Tag`]
        The tags applied to the file.
    url: Optional[:class:`str`]
        The url of this file, if given.
    """

    __slots__ = (
        "_http",
        "id",
        "created_at",
        "updated_at",
        "deletes_at",
        "favorite",
        "original_name",
        "name",
        "size",
        "type",
        "views",
        "max_views",
        "password",
        "folder_id",
        "thumbnail",
        "tags",
        "url",
    )

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    deletes_at: Optional[datetime.datetime]
    favorite: bool
    original_name: Optional[str]
    name: str
    size: int
    type: str
    views: int
    max_views: Optional[int]
    password: Optional[Union[str, bool]]
    folder_id: Optional[str]
    thumbnail: Optional[Thumbnail]  # TODO repr this object, ask diced how tf these work
    tags: Optional[List[Tag]]
    url: Optional[str]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> File:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            parse_iso_timestamp(data["deletesAt"]) if "deletesAt" in data and data["deletesAt"] is not None else None,
            data["favorite"],
            data.get("originalName"),
            data["name"],
            data["size"],
            data["type"],
            data["views"],
            data.get("maxViews"),
            data.get("password"),
            data.get("folderId"),
            data.get("thumbnail"),
            data.get("tags"),
            data.get("url"),
        )

    @property
    def full_url(self) -> str:
        """
        Returns
        -------
        :class:`str`
            The full url of this file.
        """
        if self.url and (self.url.startswith("http://") or self.url.startswith("https://")):
            url = self.url
        else:
            url = f"{self._http.base_url}{self.url}"

        return url

    @property
    def thumbnail_url(self) -> Optional[str]:
        """
        Returns
        -------
        Optional[:class:`str`]
            The full url for the thumbnail image, if available.
        """
        if not self.thumbnail:
            return None

        return f"{self._http.base_url}/{self.thumbnail.path}"

    def is_password_protected(self) -> bool:
        """
        Returns the password protection status of this file.

        Returns
        -------
        :class:`bool`
        """
        return self.password is not None

    async def download_password_protected(self, password: str) -> bytes:
        """|coro|

        Download this file if password protected.

        Parameters
        ----------
        password : :class:`str`
            The password to use when accessing this file.

        Returns
        -------
        :class:`bytes`
            The content of this file.
        """
        params = {"pw": password}
        r = Route("GET", self.full_url)
        return await self._http.request(r, params=params)

    async def refresh(self) -> File:
        """|coro|

        Retrieve an updated instance of this file.

        Returns
        -------
        :class:`~zipline.models.File`
            A new instance with the latest information about this file.
        """
        r = Route("GET", f"/api/user/files/{self.id}")
        js = await self._http.request(r)
        return File._from_data(js, http=self._http)

    async def delete(self) -> File:
        """|coro|

        Delete this file.

        Returns
        -------
        :class:`~zipline.models.File`
            A new instance with the latest information about this file.
        """
        r = Route("DELETE", f"/api/user/files/{self.id}")
        js = await self._http.request(r)
        return File._from_data(js, http=self._http)

    async def edit(
        self,
        *,
        favorite: Optional[bool] = None,
        tags: Optional[List[Tag]] = None,
        max_views: Optional[int] = None,
        original_name: Optional[str] = None,
        password: Optional[str] = None,
        type: Optional[str] = None,
    ) -> File:
        """|coro|

        Update this file.

        Parameters
        ----------
        favorite: Optional[:class:`bool`]
            The new favorite status of the file, if given.
        tags: Optional[List[:class:`~zipline.models.Tag`]]
            Tag to apply to this file, if given.
        max_views: Optional[:class:`int`]
            The new maximum views of the file, if given.
        original_name: Optional[:class:`str`]
            The new original name of the file, if given.
        password: Optional[:class:`str`]
            The new password for the file, if given.
        type: Optional[:class:`str`]
            The new MIME type for the file, if given.

            .. warning::

                Misuse of this parameter can cause files to display incorrectly or fail outright.

        Returns
        -------
        :class:`~zipline.models.File`
            The updated file.
        """
        payload = {}

        if favorite:
            payload["favorite"] = favorite
        if tags:
            payload["tags"] = [t.id for t in tags]
        if max_views:
            payload["maxViews"] = max_views
        if original_name:
            payload["originalName"] = original_name
        if password:
            payload["password"] = password
        if type:
            payload["type"] = type

        r = Route("PATCH", f"/api/user/files/{self.id}")
        js = await self._http.request(r, json=payload)
        return File._from_data(js, http=self._http)

    async def add_favorite(self) -> File:
        """|coro|

        Favorite this file.

        Returns
        -------
        :class:`~zipline.models.File`
            A new instance with the latest information about this file.

        Raises
        ------
        ValueError
            The file is already favorited.
        """
        if self.favorite:
            raise ValueError("this file is already favorited.")

        payload = {"favorite": True}
        r = Route("PATCH", f"/api/user/files/{self.id}")
        js = await self._http.request(r, json=payload)
        return File._from_data(js, http=self._http)

    async def remove_favorite(self) -> File:
        """|coro|

        Unfavorite this file.

        Returns
        -------
        :class:`~zipline.models.File`
            A new instance with the latest information about this file.

        Raises
        ------
        ValueError
            The file is already favorited.
        """
        if not self.favorite:
            raise ValueError("this file is already not favorited.")

        payload = {"favorite": False}
        r = Route("PATCH", f"/api/user/files/{self.id}")
        js = await self._http.request(r, json=payload)
        return File._from_data(js, http=self._http)

    async def remove_from_folder(self):
        """|coro|

        Remove this file from it's current :class:`~zipline.models.Folder`.
        """
        payload = {"delete": "file", "id": self.id}
        r = Route("DELETE", f"/api/user/folders/{self.folder_id}")
        await self._http.request(r, json=payload)

    async def read(self) -> bytes:
        """|coro|

        Read the content of this file into memory.

        Returns
        -------
        :class:`bytes`
            The content of this file.
        """
        if self.url and (self.url.startswith("http://") or self.url.startswith("https://")):
            url = self.url
        else:
            url = f"{self._http.base_url}{self.url}"

        r = Route("GET", url)
        return await self._http.request(r)


@dataclass
class Folder:
    """
    Represents a Folder on Zipline.

    Attributes
    ----------
    id: :class:`str`
        The id of the folder.
    created_at: :class:`datetime.datetime`
        When the folder was created.
    updated_at: :class:`datetime.datetime`
        When the folder was last updated.
    name: :class:`str`
        The name of the folder.
    public: :class:`bool`
        Whether the folder is public.
    files: Optional[List[:class:`~zipline.models.File`]]
        The files contained in the folder, if available.
    user: Optional[:class:`~zipline.models.User`]
        The user this folder belongs to, if available.
    user_id: :class:`str`
        The id of the user this folder belongs to.
    """

    __slots__ = ("_http", "id", "created_at", "updated_at", "name", "public", "files", "user", "user_id")

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    name: str
    public: bool
    files: Optional[List[File]]
    user: Optional[User]
    user_id: str

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> Folder:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            data["name"],
            data["public"],
            [File._from_data(file_data, http) for file_data in data["files"]] if "files" in data else None,  # TODO verify.
            User._from_data(data["user"], http) if "user" in data else None,
            data["userId"],
        )

    async def refresh(self) -> Folder:
        """|coro|

        Retrieve an updated instance of this object.

        Returns
        -------
        :class:`~zipline.models.Folder`
            A new instance with the latest information about this folder.
        """
        r = Route("GET", f"/api/user/folders/{self.id}")
        js = await self._http.request(r)
        return Folder._from_data(js, http=self._http)

    async def delete(self) -> Folder:
        """|coro|

        Delete this folder.

        Returns
        -------
        :class:`~zipline.models.Folder`
            A new instance with the latest information about this folder.
        """
        data = {"delete": "folder"}
        r = Route("DELETE", f"/api/user/folders/{self.id}")
        js = await self._http.request(r, json=data)
        return Folder._from_data(js, self._http)

    async def edit(self, *, name: str) -> Folder:
        """|coro|

        Edit this folder.

        Parameters
        ----------
        name: :class:`str`
            The new name of this folder.

        Returns
        -------
        :class:`~zipline.models.Folder`
            The updated folder.
        """
        data = {"name": name}
        r = Route("PATCH", f"/api/user/folders/{self.id}")
        js = await self._http.request(r, json=data)
        return Folder._from_data(js, http=self._http)

    async def remove_file(self, file: Union[File, str], /) -> Folder:
        """|coro|

        Remove a file from this folder.

        Parameters
        ----------
        file: Union[:class:`~zipline.models.File`, :class:`str`]
            The file or file id to remove from this folder.

        Returns
        -------
        :class:`~zipline.models.Folder`
            A new instance with the latest information about this folder.
        """
        payload = {
            "delete": "file",
            "id": file.id if isinstance(file, File) else file,
        }

        r = Route("DELETE", f"/api/user/folders/{self.id}")
        js = await self._http.request(r, json=payload)
        return Folder._from_data(js, http=self._http)

    async def add_file(self, file: Union[File, str], /) -> Folder:
        """|coro|

        Add a file to this folder.

        Parameters
        ----------
        file: Union[:class:`~zipline.models.File`, :class:`str`]
            The file or file id to add to this folder.

        Returns
        -------
        :class:`~zipline.models.Folder`
            A new instance with the latest information about this folder.
        """
        payload = {"id": file.id if isinstance(file, File) else file}

        r = Route("PUT", f"/api/user/folders/{self.id}")
        js = await self._http.request(r, json=payload)
        return Folder._from_data(js, http=self._http)

    async def add_files(self, files: List[Union[File, str]], /) -> Folder:
        """|coro|

        Add multiple files to this folder.

        Parameters
        ----------
        files: List[Union[:class:`~zipline.models.File`, :class:`str`]]
            The files or file ids to add to this folder.

        Returns
        -------
        :class:`~zipline.models.Folder`
            A new instance with the latest information about this folder.
        """
        file_ids = [file.id if isinstance(file, File) else file for file in files]

        payload = {
            "files": file_ids,
            "folder": self.id,
        }

        r = Route("PATCH", "/api/user/files/transaction")
        js = await self._http.request(r, json=payload)
        return Folder._from_data(js, http=self._http)


@dataclass
class User:  # TODO Maybe implement an Avatar class that decodes the b64?
    """
    Represents a Zipline user.

    Attributes
    ----------
    id: :class:`str`
        The id of this user.
    username: :class:`str`
        The username of this user.
    created_at: :class:`datetime.datetime`
        When this user was created or registered.
    updated_at: :class:`datetime.datetime`
        The last time this user was updated.
    role: :class:`~zipline.enums.UserRole`
        The role of this user.
    view: :class:`~zipline.models.UserViewSettings`
        Custom view info for this user.
    sessions: List[:class:`str`]
        List of session ids this user has open.
    oauth_providers: List[:class:`~zipline.models.OAuthProvider`]
        OAuth providers registered for this user.
    totp_secret: Optional[:class:`str`]
        The user's timed one-time password secret, if available.
    passkeys: List[:class:`~zipline.models.UserPasskey`]
        Passkeys registered for this user.
    quota: Optional[:class:`~zipline.models.UserQuota`]
        The quota this user has, if applicable.
    avatar: Optional[:class:`str`]
        The user's avatar, base64 encoded.
    password: Optional[:class:`str`]
        Password information for this user, if available.
    token: Optional[:class:`str`]
        The user's token.
    """

    __slots__ = (
        "_http",
        "id",
        "username",
        "created_at",
        "updated_at",
        "role",
        "view",
        "sessions",
        "oauth_providers",
        "totp_secret",
        "passkeys",
        "quota",
        "avatar",
        "password",
        "token",
    )

    _http: HTTPClient
    id: str
    username: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    role: UserRole
    view: Optional[UserViewSettings]
    sessions: List[str]
    oauth_providers: List[OAuthProvider]
    totp_secret: Optional[str]
    passkeys: Optional[List[UserPasskey]]
    quota: Optional[UserQuota]
    avatar: Optional[str]
    password: Optional[str]
    token: Optional[str]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> User:
        return cls(
            http,
            data["id"],
            data["username"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            UserRole(data["role"]),
            UserViewSettings._from_data(data["view"]) if "view" in data and data["view"] is not None else None,
            data["sessions"],
            data["oauthProviders"],
            data.get("totpSecret"),
            data.get("passkeys"),
            data.get("quota"),
            data.get("avatar"),
            data.get("password"),
            data.get("token"),
        )

    async def refresh(self) -> User:
        """|coro|

        Retrieve the latest information about this user.

        Returns
        -------
        :class:`~zipline.models.User`
            A new instance with the latest information about this user.
        """
        r = Route("GET", f"/api/users/{self.id}")
        js = await self._http.request(r)
        return User._from_data(js, http=self._http)

    async def edit(
        self,
        *,
        username: Optional[str] = None,
        password: Optional[str] = None,
        avatar: Optional[Tuple[str, bytes]] = None,
        role: Optional[UserRole] = None,
        quota: Optional[Union[UserQuota, PartialQuota]] = None,
    ) -> User:
        """|coro|

        Edit this user.

        Parameters
        ----------
        username: Optional[:class:`str`]
            The new username for this user, if given.
        password: Optional[:class:`str`]
            The new password for this user, if given.
        avatar: Optional[Tuple[:class:`str`, :class:`bytes`]]
            The new avatar for this user, if given.
        role: Optional[:class:`~zipline.enums.UserRole`]
            The new role for this user, if given.
        quota: Optional[Union[:class:`~zipline.models.UserQuota`, :class:`~zipline.models.PartialQuota`]]
            The new quota for this user, if given.

        Returns
        -------
        :class:`~zipline.models.User`
            The updated user.
        """
        payload = {}

        if username:
            payload["username"] = username
        if password:
            payload["password"] = password
        if avatar:
            payload["avatar"] = build_avatar_payload(*avatar)
        if role:
            payload["role"] = role.value
        if quota:
            payload["quota"] = quota._to_dict()

        r = Route("PATCH", f"/api/users/{self.id}")
        js = await self._http.request(r, json=payload)
        return User._from_data(js, http=self._http)

    async def delete(self, *, remove_data: bool = True) -> User:
        """|coro|

        Delete this user.

        Parameters
        ----------
        remove_data: :class:`bool`
            Whether this user's files and urls should be deleted as well, by default True

        Returns
        -------
        :class:`~zipline.models.User`
            A new instance with the latest information about this user.

        Raises
        ------
        BadRequest
            Something went wrong handling the request.
        Forbidden
            You are not an administrator and cannot use this method.
        """
        data = {"delete": remove_data}
        r = Route("DELETE", f"/api/users/{self.id}")
        js = await self._http.request(r, json=data)
        return User._from_data(js, http=self._http)


@dataclass
class InviteUser:
    """
    User information provided with an :class:`~zipline.models.Invite`.

    .. note::

        This can be resolved to a :class:`~zipline.models.User` via :meth:`~zipline.models.InviteUser.resolve`.

    Attributes
    ----------
    username: :class:`str`
        The username of the invite owner.
    id: :class:`str`
        The id of the invite owner.
    role: :class:`~zipline.enums.UserRole`
        The invite owner's account type.
    """

    __slots__ = ("_http", "username", "id", "role")

    _http: HTTPClient
    username: str
    id: str
    role: UserRole

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> InviteUser:
        return cls(
            http,
            data["username"],
            data["id"],
            UserRole(data["role"]),
        )

    async def resolve(self) -> User:
        """|coro|

        Resolve this object to a full :class:`~zipline.models.User`.

        Returns
        -------
        :class:`~zipline.models.User`
            The fully fledged user object.
        """
        r = Route("GET", f"/api/users/{self.id}")
        js = await self._http.request(r)
        return User._from_data(js, http=self._http)


@dataclass
class Invite:
    """
    Represents an invite to a Zipline instance.

    Attributes
    ----------
    id: :class:`str`
        The internal id of the invite.
    created_at: :class:`datetime.datetime`
        When this invite was created.
    updated_at: :class:`datetime.datetime`
        When this invite was last updated.
    expires_at: Optional[:class:`datetime.datetime`]
        When this invite expires, if applicable.
    code: :class:`str`
        The code for this invite.
    uses: :class:`int`
        The number of times this invite has been used.
    max_uses: Optional[:class:`int`]
        The number of times this invite can be used before it's no longer valid, if applicable.
    inviter: :class:`~zipline.models.InviteUser`
        The user that is attributed to the creation of this invite.
    inviter_id: :class:`str`
        The id of the user attributed to the creation of this invite.
    """

    __slots__ = (
        "_http",
        "id",
        "created_at",
        "updated_at",
        "expires_at",
        "code",
        "uses",
        "max_uses",
        "inviter",
        "inviter_id",
    )

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    expires_at: Optional[datetime.datetime]
    code: str
    uses: int
    max_uses: Optional[int]
    inviter: InviteUser
    inviter_id: str

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> Invite:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            parse_iso_timestamp(data["expiresAt"]) if "expiresAt" in data and data["expiresAt"] is not None else None,
            data["code"],
            data["uses"],
            data.get("maxUses"),
            InviteUser._from_data(data["inviter"], http=http),
            data["inviterId"],
        )

    @property
    def url(self) -> str:
        """
        Returns
        -------
        :class:`str`
            The full url of this invite.
        """
        return f"{self._http.base_url}/auth/register?code={self.code}"

    async def delete(self) -> Invite:
        """|coro|

        Delete this invite.

        Returns
        -------
        :class:`~zipline.models.Invite`
            A new instance with the latest information about this invite.
        """
        r = Route("DELETE", f"/api/auth/invites/{self.id}")
        js = await self._http.request(r)
        return Invite._from_data(js, self._http)

    async def refresh(self) -> Invite:
        """|coro|

        Retreive updated information about this invite.

        Returns
        -------
        :class:`~zipline.models.Invite`
            A new instance with the latest information about this invite.
        """
        r = Route("GET", f"/api/auth/invites/{self.id}")
        js = await self._http.request(r)
        return Invite._from_data(js, http=self._http)


class TagFile:
    """
    Partial file given with Tags.

    .. note::

        This can be resolved to a :class:`~zipline.models.File` via :meth:`~zipline.models.TagFile.resolve`.

    Attributes
    ----------
    id: :class:`str`
        The internal id of the :class:`~zipline.models.File` represented.
    """

    __slots__ = ("_http", "id")

    def __init__(self, http: HTTPClient, id: str):
        self._http = http
        self.id = id

    async def resolve(self) -> File:
        """|coro|

        Retrieve the whole file associated with this partial data.

        Returns
        -------
        :class:`~zipline.models.File`
            The fully fledged file object.
        """
        r = Route("GET", f"/api/user/files/{self.id}")  # NOTE: Undocumented // Not officially used in the frontend.
        js = await self._http.request(r)
        return File._from_data(js, http=self._http)


@dataclass
class Tag:
    """
    Represents a tag on Zipline.

    Attributes
    ----------
    id: :class:`str`
        The internal id of the tag.
    created_at: :class:`datetime.datetime`
        When this tag was created.
    updated_at: :class:`datetime.datetime`
        When this tag was last updated.
    name: :class:`str`
        The name of this tag.
    color: :class:`str`
        The color associated with this tag.
    files: Optional[List[:class:`~zipline.models.TagFile`]]
        Partial files associated with this tag, if available.
    """

    __slots__ = ("_http", "id", "created_at", "updated_at", "name", "color", "files")

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    name: str
    color: str
    files: Optional[List[TagFile]]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> Tag:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            data["name"],
            data["color"],
            [TagFile(http, r["id"]) for r in data["files"]] if "files" in data else None,
        )

    async def edit(self, color: Optional[str] = None, name: Optional[str] = None) -> Tag:
        """|coro|

        Edit this tag.

        Parameters
        ----------
        color: Optional[:class:`str`]
            The new color of the tag, if given.

            .. note::

                Must be in hex format with preceeding #

                ex. #rrggbb
        name: Optional[:class:`str`]
            The new name of the tag, if given.

        Returns
        -------
        :class:`~zipline.models.Tag`
            The updated tag.
        """
        payload = {}

        if color:
            payload["color"] = color
        if name:
            payload["name"] = name

        r = Route("PATCH", f"/api/user/tags/{self.id}")
        js = await self._http.request(r, json=payload)
        return Tag._from_data(js, http=self._http)

    async def delete(self) -> Tag:
        """|coro|

        Delete this tag.

        Returns
        -------
        :class:`~zipline.models.Tag`
            A new instance with the latest information about this tag.
        """
        r = Route("DELETE", f"/api/user/tags/{self.id}")
        js = await self._http.request(r)
        return Tag._from_data(js, http=self._http)

    async def refresh(self) -> Tag:
        """|coro|

        Retrieve the latest information about this tag.

        Returns
        -------
        :class:`~zipline.models.Tag`
            A new instance with the latest information about this tag.
        """
        r = Route("GET", f"/api/user/tags/{self.id}")
        js = await self._http.request(r)
        return Tag._from_data(js, http=self._http)


@dataclass
class URL:
    """
    Represents a shortened url on Zipline.

    Attributes
    ----------
    id: :class:`str`
        The internal id of the url.
    created_at: :class:`datetime.datetime`
        When this url was created.
    updated_at: :class:`datetime.datetime`
        When this url was last updated.
    code: :class:`str`
        The url code.
    vanity: Optional[:class:`str`]
        The vanity code of this url, if applicable.
    destination: :class:`str`
        The url this url redirects to.
    views: :class:`int`
        The number of times this url has been used.
    max_views: Optional[:class:`int`]
        The maximum number of times this url can be used, if applicable.
    password: Optional[:class:`str`]
        The password required to use this url.
    enabled: :class:`bool`
        Whether this url is active for use.
    user: Optional[:class:`~zipline.models.User`]
        The user this url belongs to.
    user_id: :class:`str`
        The id of the user this url belongs to.
    """

    __slots__ = (
        "_http",
        "id",
        "created_at",
        "updated_at",
        "code",
        "vanity",
        "destination",
        "views",
        "max_views",
        "password",
        "enabled",
        "user",
        "user_id",
    )

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    code: str  # The part of the url that redirects. <base_url>/go/<code>
    vanity: Optional[str]
    destination: str
    views: int
    max_views: Optional[int]
    password: Optional[str]
    enabled: bool
    user: Optional[User]
    user_id: str

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> URL:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            data["code"],
            data.get("vanity"),
            data["destination"],
            data["views"],
            data.get("maxViews"),
            data.get("password"),
            data["enabled"],
            User._from_data(data["user"], http) if "user" in data else None,
            data["userId"],
        )

    @property
    def full_url(self) -> str:
        """
        Returns
        -------
        :class:`str`
            The full url.
        """
        code = self.vanity or self.code
        return f"{self._http.base_url}/go/{code}"

    async def edit(
        self,
        max_views: Optional[int] = None,
        vanity: Optional[str] = None,
        destination: Optional[str] = None,
        enabled: Optional[bool] = None,
        password: Optional[str] = None,
    ) -> URL:
        """|coro|

        Edit this url.

        Parameters
        ----------
        max_views: Optional[:class:`int`]
            The new maximum views, if given.
        vanity: Optional[:class:`str`]
            The new vanity code this url should have, if given.
        destination: Optional[:class:`str`]
            The new destination, if given.
        enabled: Optional[:class:`bool`]
            Whether this url should be usable, if given.
        password: Optional[:class:`str`]
            The new password required to use this url, if given.

        Returns
        -------
        :class:`~zipline.models.URL`
            The updated url.
        """
        payload = {}

        if max_views:
            payload["maxViews"] = max_views
        if vanity:
            payload["vanity"] = vanity
        if destination:
            payload["destination"] = destination
        if enabled:
            payload["enabled"] = enabled
        if password:
            payload["password"] = password

        r = Route("PATCH", f"/api/user/urls/{self.id}")
        js = await self._http.request(r, json=payload)
        return URL._from_data(js, http=self._http)

    async def delete(self) -> URL:
        """|coro|

        Delete this url.

        Returns
        -------
        :class:`~zipline.models.URL`
            A new instance with the latest information about this url.
        """
        r = Route("DELETE", f"/api/user/urls/{self.id}")
        js = await self._http.request(r)
        return URL._from_data(js, http=self._http)


@dataclass
class UploadFile:
    """
    File data given by the API when a file is uploaded.

    .. note::

        This may be resolved to a full :class:`~zipline.models.File` via :meth:`~zipline.models.UploadFile.resolve`.

    Attributes
    ----------
    id: :class:`str`
        The id of the uploaded file.
    type: :class:`str`
        The media type of the uploaded file.
    url: :class:`str`
        The url for the uploaded file.
    """

    __slots__ = ("_http", "id", "type", "url")

    _http: HTTPClient
    id: str
    type: str
    url: str

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> UploadFile:
        return cls(
            http,
            data["id"],
            data["type"],
            data["url"],
        )

    async def resolve(self) -> File:
        """|coro|

        Retrieve fully fledged information about this object.

        Returns
        -------
        :class:`~zipline.models.File`
            The fully fledged object
        """
        r = Route("GET", f"/api/user/files/{self.id}")  # NOTE: Undocumented // Not officially used in the frontend.
        js = await self._http.request(r)
        return File._from_data(js, self._http)


@dataclass
class UploadResponse:
    """
    Response given by the API when a file or multiple files are uploaded.

    Attributes
    ----------
    files: List[:class:`~zipline.models.UploadFile`]
        A list of information about the files uploaded.
    """

    __slots__ = ("_http", "files", "deletes_at")

    _http: HTTPClient
    files: List[UploadFile]
    deletes_at: Optional[datetime.datetime]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> UploadResponse:
        return cls(
            http,
            [UploadFile._from_data(uf, http) for uf in data["files"]],
            parse_iso_timestamp(data["deletesAt"]) if "deletesAt" in data else None,
        )


class FileData:
    """
    Used to upload a File to Zipline.

    Attributes
    ----------
    data: Union[:class:`str`, :class:`bytes`, :class:`os.PathLike`, :class:`io.BufferedIOBase`]
        The file or file like object to open.
    filename: Optional[:class:`str`]
        The name of the file to be uploaded. Defaults to filename of the given path, if applicable.
    mimetype: Optional[:class:`str`]
        The MIME type of the file, if None the lib will attempt to determine it.
    """

    __slots__ = ("filename", "data", "mimetype")

    def __init__(
        self,
        data: Union[str, bytes, os.PathLike[Any], io.BufferedIOBase],
        filename: Optional[str] = None,
        *,
        mimetype: Optional[str] = None,
    ) -> None:
        """Used to upload a file to Zipline.

        Parameters
        ----------
        data: Union[:class:`str`, :class:`bytes`, :class:`os.PathLike`, :class:`io.BufferedIOBase`]
            The file or file like object to open.
        filename: Optional[:class:`str`]
            The name of the file to be uploaded. Defaults to filename of the given path, if applicable.
        mimetype: Optional[:class:`str`]
            The MIME type of the file, if None the library will attempt to determine it.

        Raises
        ------
        ValueError
            An invalid value was passed to the data parameter.
        TypeError
            The MIME type of the file could not be determined and was not provided.
        """
        if isinstance(data, io.IOBase):
            if not (data.seekable() and data.readable()):
                raise ValueError(f"File buffer {data!r} must be seekable and readable")
            self.data: io.BufferedIOBase = data
        else:
            self.data = open(data, "rb")

        # Mime type determination. Strategy is:
        #   - an explicitly given type
        #   - a guessed type based on the filename, if present.
        #   - a guessed type based on magic bytes
        #   - application/octet-stream as a fallback.
        if mimetype is not None:
            guessed_mime = mimetype
        elif filename is not None:
            guessed_mime = mimetypes.guess_type(filename)[0]
        elif filename is None and mimetype is None:
            guessed_mime = guess_mimetype_by_magicnumber(self.data.read(16))
            # back it up again
            self.data.seek(0)

        self.mimetype = guessed_mime or "application/octet-stream"

        if filename is None:
            if isinstance(data, str):
                _, filename = os.path.split(data)
            else:
                filename = getattr(data, "name", "untitled")

        self.filename = filename

        if self.mimetype is None:
            raise TypeError("could not determine mimetype of file given")


class PartialQuota:
    """
    A partial quota useful for creating a new quota for a :class:`~zipline.models.User`.

    Attributes
    ----------
    type: :class:`~zipline.enums.QuotaType`
        The type of this quota.
    value: Optional[:class:`int`]
        The value to assign to this quota.

        .. note::

            This may only be omitted if :attr:`~zipline.PartialQuota.type` is :attr:`~zipline.QuotaType.none`
    max_urls: Optional[int]
        The url limit for this quota.

        .. note::

            This argument may be omitted. If None is passed there will be no limit.
    """

    def __init__(self, type: QuotaType, value: Optional[int] = None, max_urls: Optional[int] = MISSING):
        if self.value and self.value < 0:
            raise ValueError("quota amount must be greater than or equal to zero.")

        if self.type is not QuotaType.none and self.value is None:
            raise ValueError("value must be given to this quota unless type is none.")

        if isinstance(max_urls, int) and max_urls <= 0:
            raise ValueError("max_urls must be greater than or equal to zero.")

        self.type = type
        self.value = value
        self.max_urls = max_urls

    def _to_dict(self) -> Dict[str, Any]:
        return generate_quota_payload(self.type, self.value, self.max_urls)


@dataclass
class UserQuota:
    """
    Represents a quota assigned to a :class:`~zipline.models.User` in Zipline.

    .. note::

        While this class may be used to update a user's quota this class is not intended
        to be created manually and should not be used for this purpose.

        Usage of :class:`~zipline.models.PartialQuota` is recommended this purpose instead.

    Attributes
    ----------
    id: :class:`str`
        The id of the quota.
    created_at: :class:`datetime.datetime`
        When this quota was created.
    updated_at: :class:`datetime.datetime`
        When this quota was last updated.
    files_quota: :class:`~zipline.enums.QuotaType`
        The type of this quota.
    max_bytes: Optional[:class:`str`]
        The maximum bytes of storage for this quota, if applicable.
    max_files: Optional[:class:`int`]
        The maximum number of files for this quota, if applicable.
    max_urls: Optional[:class:`int`]
        The maximum number of shortened urls for this quota, if applicable.
    user: Optional[:class:`~zipline.models.User`]
        The user this quota is assigned to, if given.
    user_id: Optional[:class:`str`]
        The id of the user this quota is assigned to, if given.
    """

    __slots__ = (
        "_http",
        "id",
        "created_at",
        "updated_at",
        "files_quota",
        "max_bytes",
        "max_files",
        "max_urls",
        "user",
        "user_id",
    )

    _http: HTTPClient
    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    files_quota: QuotaType
    max_bytes: Optional[str]
    max_files: Optional[int]
    max_urls: Optional[int]
    user: Optional[User]
    user_id: Optional[str]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /, http: HTTPClient) -> UserQuota:
        return cls(
            http,
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            QuotaType(data["filesQuota"]),
            data.get("maxBytes"),
            data.get("maxFiles"),
            data.get("maxUrls"),
            User._from_data(data["user"], http=http) if "user" in data else None,
            data.get("userId"),
        )

    def _to_dict(self) -> Dict[str, Any]:
        return generate_quota_payload(self.type, self._amount(), self.max_urls)

    def _amount(self) -> int:
        if self.max_bytes is not None:
            return int(self.max_bytes)

        if self.max_files is not None:
            return self.max_files

        raise ValueError("amount of this quota cannot be determined.")

    @property
    def type(self) -> QuotaType:
        """
        Returns
        -------
        :class:`~zipline.enums.QuotaType`
            The type of this quota.
        """
        return self.files_quota

    async def resolve_user(self) -> User:
        """|coro|

        Resolve the user this quota is assigned to, if possible.

        Returns
        -------
        :class:`~zipline.models.User`
            The user this quota is assigned to.

        Raises
        ------
        TypeError
            The :attr:`Quota.user_id` is None and the user cannot be resolved.
        """
        if self.user_id is None:
            raise TypeError("cannot resolve user with null id.")

        r = Route("GET", f"/api/users/{self.id}")
        js = await self._http.request(r)
        return User._from_data(js, http=self._http)


@dataclass
class UserPasskey:
    """
    A passkey a Zipline :class:`~zipline.models.User` has set up.

    Attributes
    ----------
    id: :class:`str`
        The internal id of this object.
    created_at: :class:`datetime.datetime`
        When this passkey was created in the database.
    updated_at: :class:`datetime.datetime`
        When this passkey was last updated.
    last_used: Optional[:class:`datetime.datetime`]
        When this passkey was last used.
    name: :class:`str`
        The name of this passkey.
    reg: :class:`dict`
        Additional information about this passkey.
    user_id: :class:`str`
        The id of the :class:`~zipline.models.User` that this passkey belongs to.
    """

    __slots__ = ("id", "created_at", "updated_at", "last_used", "name", "reg", "user_id")

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    last_used: Optional[datetime.datetime]
    name: str
    reg: JSON
    user_id: str  # NOTE: This does expose a `User` attr in the api.
    #                     Not inclined to expose it because it'd complicate this class
    #                     for minimal gain as this should already be attached to a User object.

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /) -> UserPasskey:
        return cls(
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            parse_iso_timestamp(data["lastUsed"]) if "lastUsed" in data else None,
            data["name"],
            data["reg"],
            data["userId"],
        )


@dataclass
class OAuthProvider:
    """
    Represents an OAuth provider being used by a :class:`~zipline.models.User` on Zipline.

    Attributes
    ----------
    id: :class:`str`
        The internal id of this entry.
    created_at: :class:`datetime.datetime`
        When this entry was created.
    updated_at: :class:`datetime.datetime`
        When this entry was updated.
    user_id: :class:`str`
        The id of the user this provider registration is for.
    provider: :class:`~zipline.enums.OAuthProviderType`
        The service this entry is for.
    username: :class:`str`
        The username of this entry.
    access_token: :class:`str`
        The access token of this entry.
    refresh_token: Optional[:class:`str`]
        The refresh token for this entry.
    oauth_id: Optional[:class:`str`]
        The oauth id for this entry.
    """

    __slots__ = (
        "id",
        "created_at",
        "updated_at",
        "user_id",
        "provider",
        "username",
        "access_token",
        "refresh_token",
        "oauth_id",
    )

    id: str
    created_at: datetime.datetime
    updated_at: datetime.datetime
    user_id: str
    provider: OAuthProviderType
    username: str
    access_token: str
    refresh_token: Optional[str]
    oauth_id: Optional[str]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /) -> OAuthProvider:
        return cls(
            data["id"],
            parse_iso_timestamp(data["createdAt"]),
            parse_iso_timestamp(data["updatedAt"]),
            data["userId"],
            OAuthProviderType(data["provider"]),
            data["username"],
            data["accessToken"],
            data.get("refreshToken"),
            data.get("oauthId"),
        )


class Thumbnail:
    """
    Thumbnail data for a Zipline :class:`~zipline.models.File`.

    Attributes
    ----------
    path: :class:`str`
        The path to this thumbnail.
    """

    __slots__ = ("path",)

    def __init__(self, path: str):
        self.path = path


@dataclass
class UserViewSettings:
    """
    Represents view settings for a Zipline :class:`~zipline.models.User`.

    Attributes
    ----------
    enabled: Optional[:class:`bool`]
        Whether the view settings are enabled.
    align: Optional[Literal['left', 'center', 'right']]
        How the content should be aligned on the page.
    show_mimetype: Optional[:class:`bool`]
        Whether the content's MIME type should be displayed.
    content: Optional[:class:`str`]
        The view content.
    embed: Optional[:class:`bool`]
        Whether embed tags should be present.
    embed_title: Optional[:class:`str`]
        The title of the embed the content will generate, if applicable.
    embed_description: Optional[:class:`str`]
        The description of the embed the content will generate, if applicable.
    embed_color: Optional[:class:`str`]
        The color of the embed the content will generate, if applicable.
    embed_site_name: Optional[:class:`str`]
        The name of the site the embed will redirect to, if applicable.
    """

    __slots__ = (
        "enabled",
        "align",
        "show_mimetype",
        "content",
        "embed",
        "embed_title",
        "embed_description",
        "embed_color",
        "embed_site_name",
    )

    enabled: Optional[bool]
    align: Optional[Literal["left", "center", "right"]]
    show_mimetype: Optional[bool]
    content: Optional[str]
    embed: Optional[bool]
    embed_title: Optional[str]
    embed_description: Optional[str]
    embed_color: Optional[str]
    embed_site_name: Optional[str]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /) -> UserViewSettings:
        return cls(
            data.get("enabled"),
            data.get("align"),
            data.get("showMimetype"),
            data.get("content"),
            data.get("embed"),
            data.get("embedTitle"),
            data.get("embedDescription"),
            data.get("embedColor"),
            data.get("embedSiteName"),
        )


@dataclass
class ServerVersionInfo:
    """
    Version information for a Zipline instance.

    Attributes
    ----------
    version: :class:`str`
        The current version being used.
    """

    __slots__ = ("version",)

    version: str

    @classmethod
    def _from_data(cls, data: Dict[str, Any]) -> ServerVersionInfo:
        return cls(data["version"])


@dataclass
class UserStats:
    """
    Stats for a Zipline :class:`~zipline.models.User`.

    Attributes
    ----------
    files_uploaded: :class:`int`
        The number of files the user has uploaded.
    favorite_files: :class:`int`
        The number of files the user has favorited.
    views: :class:`int`
        The number of times files uploaded by the user have been viewed.
    avg_views: :class:`int`
        The average number of views a file uploaded by the user has received.
    storage_used: :class:`int`
        The amount of storage the user is using in bytes.
    avg_storage_used: :class:`float`
        The amount of storage a file uploaded by the user takes up in bytes.
    urls_created: :class:`int`
        The number of urls the user has created.
    url_views: :class:`int`
        The number of times urls created by the user have been visited.
    sort_type_count: Dict[:class:`str`, :class:`int`]
        A mapping of MIME type to number of files uploaded.
    """

    __slots__ = (
        "files_uploaded",
        "favorite_files",
        "views",
        "avg_views",
        "storage_used",
        "avg_storage_used",
        "urls_created",
        "url_views",
        "sort_type_count",
    )

    files_uploaded: int
    favorite_files: int
    views: int
    avg_views: int
    storage_used: int
    avg_storage_used: float
    urls_created: int
    url_views: int
    sort_type_count: Dict[str, int]

    @classmethod
    def _from_data(cls, data: Dict[str, Any]) -> UserStats:
        return cls(
            data["filesUploaded"],
            data["favoriteFiles"],
            data["views"],
            data["avgViews"],
            data["storageUsed"],
            data["avgStorageUsed"],
            data["urlsCreated"],
            data["urlViews"],
            data["sortCountType"],
        )


@dataclass
class UserFilesResponse:
    """
    Represents a response to a file search on Zipline.

    Attributes
    ----------
    page: List[:class:`~zipline.models.File`]
        The files on this page.
    search: Optional[Dict[:class:`str`, Any]]
        Search parameters that were passed.
    total: Optional[:class:`int`]
        The total number of items matching the search.
    pages: Optional[:class:`int`]
        The number of pages available.
    """

    __slots__ = ("page", "search", "total", "pages")
    page: List[File]
    search: Optional[Dict[str, Any]]
    total: Optional[int]
    pages: Optional[int]

    @classmethod
    def _from_data(cls, data: Dict[str, Any], /) -> UserFilesResponse:
        return cls(
            data["page"],
            data.get("search"),
            data.get("total"),
            data.get("pages"),
        )



"""
Copyright 2023-present fretgfr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import base64
import datetime
from itertools import islice
from operator import attrgetter
from typing import (
    Any,
    AsyncIterable,
    Callable,
    Coroutine,
    Dict,
    Generator,
    Iterable,
    Optional,
    Tuple,
    TypeVar,
    Union,
    overload,
)

from .enums import QuotaType

__all__ = (
    "as_chunks",
    "get",
    "utcnow",
)

T = TypeVar("T")
_Iter = Union[Iterable[T], AsyncIterable[T]]
Coro = Coroutine[Any, Any, T]


def parse_iso_timestamp(iso_str: str, /) -> datetime.datetime:
    """Parses an ISO string to an aware UTC datetime."""
    return datetime.datetime.strptime(iso_str, "%Y-%m-%dT%H:%M:%S.%fZ").replace(tzinfo=datetime.timezone.utc)


def to_iso_format(dt: datetime.datetime, /) -> str:
    """Transforms a datetime.datetime to an ISO string."""
    return dt.strftime("%Y-%m-%dT%H:%M:%S.%fZ")


def utcnow() -> datetime.datetime:
    """Returns an aware UTC datetime representing the current time.

    Returns
    --------
    :class:`datetime.datetime`
        The current time in UTC as an aware datetime.
    """
    return datetime.datetime.now(datetime.timezone.utc)


# From the python docs for 3.12s `itertools.batched`
def as_chunks(iterable: Iterable[T], n: int) -> Generator[Tuple[T, ...], None, None]:
    """Batches an iterable into chunks of up to size n.

    Parameters
    ----------
    iterable: :class:`collections.abc.Iterable`
        The iterable to batch
    n: :class:`int`
        The number of elements per generated tuple.

    Raises
    ------
    :class:`ValueError`
        At least one result must be returned per group.
    """
    if n < 1:
        raise ValueError("n must be at least one")
    it = iter(iterable)
    while batch := tuple(islice(it, n)):
        yield batch


# The following get utility was sourced from discord.py under the MIT license.
def _get(iterable: Iterable[T], /, **attrs: Any) -> Optional[T]:
    # global -> local
    _all = all
    attrget = attrgetter

    # Special case the single element call
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrget(k.replace("__", "."))
        return next((elem for elem in iterable if pred(elem) == v), None)

    converted = [(attrget(attr.replace("__", ".")), value) for attr, value in attrs.items()]
    for elem in iterable:
        if _all(pred(elem) == value for pred, value in converted):
            return elem
    return None


async def _aget(iterable: AsyncIterable[T], /, **attrs: Any) -> Optional[T]:
    # global -> local
    _all = all
    attrget = attrgetter

    # Special case the single element call
    if len(attrs) == 1:
        k, v = attrs.popitem()
        pred = attrget(k.replace("__", "."))
        async for elem in iterable:
            if pred(elem) == v:
                return elem
        return None

    converted = [(attrget(attr.replace("__", ".")), value) for attr, value in attrs.items()]

    async for elem in iterable:
        if _all(pred(elem) == value for pred, value in converted):
            return elem
    return None


@overload
def get(iterable: AsyncIterable[T], /, **attrs: Any) -> Coro[Optional[T]]: ...


@overload
def get(iterable: Iterable[T], /, **attrs: Any) -> Optional[T]: ...


def get(iterable: _Iter[T], /, **attrs: Any) -> Union[Optional[T], Coro[Optional[T]]]:
    r"""A helper that returns the first element in the iterable that meets
    all the traits passed in ``attrs``.

    When multiple attributes are specified, they are checked using
    logical AND, not logical OR. Meaning they have to meet every
    attribute passed in and not one of them.

    To have a nested attribute search (i.e. search by ``x.y``) then
    pass in ``x__y`` as the keyword argument.

    If nothing is found that matches the attributes passed, then
    ``None`` is returned.


    Parameters
    -----------
    iterable: Union[:class:`collections.abc.Iterable`, :class:`collections.abc.AsyncIterable`]
        The iterable to search through. Using a :class:`collections.abc.AsyncIterable`,
        makes this function return a coroutine.
    \*\*attrs
        Keyword arguments that denote attributes to search with.
    """

    return (
        _aget(iterable, **attrs)  # type: ignore
        if hasattr(iterable, "__aiter__")  # isinstance(iterable, collections.abc.AsyncIterable) is too slow
        else _get(iterable, **attrs)  # type: ignore
    )


def safe_get(dict_: Dict[Any, Any], *keys: Any) -> Optional[Any]:
    """Safely get a nested key from a dictionary.

    ex.
    .. code-block:: python3

        d = {
            'i': 1,
            'j': {
                'k': '2'
            }
        }

        safe_get(d, 'i')  # -> 1
        safe_get(d, 'j', 'k')  # -> '2'


    Parameters
    ----------
    dict_ : Dict[Any, Any]
        The dictionary to get the key from
    keys : Any
        The keys to get from the dictionary. Processed in order.

    Returns
    -------
    Optional[Any]
        The value at the key. None if the key could not be gotten.
    """
    for key in keys:
        try:
            dict_ = dict_[key]
        except KeyError:
            return None

    return dict_


def build_avatar_payload(mime: str, data: bytes) -> str:
    """
    Builds an encoded string that can be used to upload avatar data.

    Parameters
    ----------
    mime: str
        The MIME of the data given.
    data: bytes
        The payload's actual data.

    Returns
    -------
    str
        The encoded data
    """
    avatar_bytes_encoded = base64.b64encode(data).decode("ascii")

    return f"data:{mime};base64,{avatar_bytes_encoded}"


def copy_doc(original: Callable[..., Any]) -> Callable[[T], T]:
    def decorator(overridden: T) -> T:
        overridden.__doc__ = original.__doc__
        overridden.__signature__ = _signature(original)  # type: ignore
        return overridden

    return decorator


class _MissingSentinel:
    __slots__ = ()

    def __eq__(self, other) -> bool:
        return False

    def __bool__(self) -> bool:
        return False

    def __hash__(self) -> int:
        return 0

    def __repr__(self):
        return "..."


MISSING: Any = _MissingSentinel()


def generate_quota_payload(
    type: QuotaType, value: Optional[int] = None, max_urls: Optional[int] = MISSING
) -> Dict[str, Any]:
    payload: Dict[str, Any] = {
        "filesType": type.value,
    }

    if type is QuotaType.by_bytes:
        payload["maxBytes"] = str(value)
    elif type is QuotaType.by_files:
        payload["maxFiles"] = value
    elif type is QuotaType.none:
        payload["maxBytes"] = None
        payload["maxFiles"] = None

    if max_urls is not MISSING:
        payload["maxUrls"] = max_urls

    return payload


def guess_mimetype_by_magicnumber(data: bytes) -> Optional[str]:
    """
    Sourced: https://en.wikipedia.org/wiki/List_of_file_signatures

    Handled types:
        - jpeg
        - png
        - webp
        - gif
        - mp4
        - mkv
        - mp3
        - mov
    """
    if data[0:3] == b"\xff\xd8\xff" or data[6:10] in (b"JFIF", b"Exif"):
        return "image/jpeg"
    elif data.startswith(b"\x89\x50\x4E\x47\x0D\x0A\x1A\x0A"):
        return "image/png"
    elif data.startswith(b"RIFF") and data[8:12] == b"WEBP":
        return "image/webp"
    elif data.startswith((b"\x47\x49\x46\x38\x37\x61", b"\x47\x49\x46\x38\x39\x61")):
        return "image/gif"
    elif data[3:11] in (b"\x66\x74\x79\x70\x4d\x53\x4e\x56", b"\x66\x74\x79\x70\x69\x73\x6f\x6d"):
        return "video/mp4"
    elif data[0:4] == b"\x1a\x45\xdf\xa3":
        return "video/x-matroska"
    elif data[0:2] in (b"\xff\xfb", b"\xff\xf3", b"\xff\xf2") or data[0:3] == b"\x49\x44\x43":
        return "audio/mpeg"
    elif data[0:4] == b"\x6d\x6f\x6f\x76":
        return "video/quicktime"
    else:
        return None



"""
Copyright 2023-present fretgfr

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import datetime
from types import TracebackType
from typing import TYPE_CHECKING, Dict, List, Optional, Tuple, Type, Union

import aiohttp

from .enums import FileSearchField, FileSearchSort, NameFormat, Order, RecentFilesFilter
from .http import HTTPClient, Route
from .models import (
    URL,
    File,
    FileData,
    Folder,
    Invite,
    ServerVersionInfo,
    Tag,
    UploadResponse,
    User,
    UserFilesResponse,
    UserRole,
    UserStats,
)
from .utils import build_avatar_payload, to_iso_format, utcnow

if TYPE_CHECKING:
    from typing_extensions import Self

__all__ = ("Client",)


class Client:
    """A Zipline Client.

    .. container:: operations

        .. describe:: async with x:

            Returns the Client itself. Used to gracefully close the client on exit.

            .. code-block:: python3

                async with zipline.Client(server_url, token) as client:
                    ...
    """

    __slots__ = ("server_url", "http")

    def __init__(self, server_url: str, token: str) -> None:
        """Creates a new Client.

        Parameters
        ----------
        server_url: :class:`str`
            The URL of the Zipline server.
        token: :class:`str`
            Your Zipline token.
        """
        self.server_url = server_url
        self.http = HTTPClient(server_url, token)

    async def get_version(self) -> ServerVersionInfo:
        """|coro|

        Gets the Zipline server version information.

        Returns
        -------
        :class:`~zipline.models.ServerVersionInfo`
            The version information for the server.
        """
        r = Route("GET", "/api/version")
        js = await self.http.request(r)
        return ServerVersionInfo._from_data(js)

    async def get_user_stats(self) -> UserStats:
        """|coro|

        Retrieve stats about the current user.

        Returns
        -------
        :class:`~zipline.models.UserStats`
            Stats for the current user.
        """
        r = Route("GET", "/api/user/stats")
        js = await self.http.request(r)
        return UserStats._from_data(js)

    async def create_user(
        self,
        *,
        username: str,
        password: str,
        role: UserRole = UserRole.user,
        avatar: Optional[Tuple[str, bytes]] = None,
    ) -> User:
        """|coro|

        Creates a user.

        Parameters
        ----------
        username: :class:`str`
            The username of the user to create.
        password: :class:`str`
            The password of the user to create.
        role: :class:`~zipline.enums.UserRole`
            The permissions level of the created User. Only available if used by a Super Admin.
        avatar: Optional[Tuple[:class:`str`, :class:`bytes`]]
            If given, a tuple containing a string denoting the MIME type of the data being uploaded and the data itself as bytes.

            Example:

            .. code:: python3

                with open('avatar.png', 'rb') as fp:
                    avatar_bytes = fp.read()

                avatar_mime = 'image/png'

                avatar = (avatar_mime, avatar_bytes)

                user = await create_user(
                    username='Example',
                    password='s0m3th!ngs3cur3',
                    avatar=avatar,
                )

        Returns
        -------
        :class:`~zipline.models.User`
            The created user.

        Raises
        ------
        BadRequest
            Something went wrong handling the request.
        Forbidden
            You are not an administrator and cannot use this method.
        """
        json = {"username": username, "password": password, "role": role.value}

        if avatar:
            json["avatar"] = build_avatar_payload(*avatar)

        r = Route("POST", "/api/users")
        data = await self.http.request(r, json=json)
        return User._from_data(data, http=self.http)

    async def get_user(self, id: str, /) -> User:
        """|coro|

        Retreive a user with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user to get.

        Returns
        -------
        :class:`~zipline.models.User`
            The retrieved user.

        Raises
        ------
        Forbidden
            You are not an administrator and cannot use this method.
        NotFound
            A user with that id could not be found
        """
        r = Route("GET", f"/api/users/{id}")
        js = await self.http.request(r)
        return User._from_data(js, http=self.http)

    async def get_all_users(self) -> List[User]:
        """|coro|

        Gets all users.

        Returns
        -------
        List[:class:`~zipline.models.User`]
            The retrieved users

        Raises
        ------
        Forbidden
            You are not an administrator and cannot use this method.
        """
        r = Route("GET", "/api/users")
        js = await self.http.request(r)
        return [User._from_data(data, http=self.http) for data in js]

    async def delete_user(self, id, /, *, remove_data: bool = True) -> User:
        """|coro|

        Delete a user with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the user to delete.
        remove_data: :class:`bool`
            Whether the user's files and urls should be removed, by default True

        Returns
        -------
        :class:`~zipline.models.User`
            The deleted user.

        Raises
        ------
        BadRequest
            Something went wrong handling the request.
        Forbidden
            You are not an administrator and cannot use this method.
        """
        data = {"delete": remove_data}
        r = Route("DELETE", f"/api/users/{id}")
        js = await self.http.request(r, json=data)
        return User._from_data(js, http=self.http)

    async def get_all_invites(self) -> List[Invite]:
        """|coro|

        Retrieves all invites.

        .. warning::

            This method may retrieve invites for all users, not just the user the token belongs to.

        Returns
        -------
        List[:class:`~zipline.models.Invite`]
            The invites on the server.

        Raises
        ------
        BadRequest
            Something went wrong handling this request.
        Forbidden
            You are not an administrator and do not have permission to access this resource.
        """
        r = Route("GET", "/api/auth/invites")
        js = await self.http.request(r)
        return [Invite._from_data(data, http=self.http) for data in js]

    async def create_invite(
        self,
        *,
        max_uses: Optional[int] = 1,
        expires_at: Optional[Union[datetime.datetime, datetime.timedelta]] = None,
    ) -> Invite:
        """|coro|

        Creates an invite.

        .. versionchanged:: 0.21.0

            Removed the count parameter as it's no longer supported.
            Return type is now a complete Invite object as opposed to a partial model.


        Parameters
        ----------
        max_uses: Optional[:class:`int`]
            The number of times the created invite can be used, by default 1.
            If None is passed the invite will have unlimited uses.

            .. versionadded:: 0.21.0
        expires_at: Optional[:class:`datetime.datetime`]
            When the created invite(s) should expire. Defaults to 24 hours from creation.

            .. versionchanged:: 0.21.0

                Default expiration removed.
                This parameter will now accept a timedelta.
                If None is passed, the invite will never expire.

        Returns
        -------
        :class:`~zipline.models.Invite`
            The created invite.

        Raises
        ------
        ValueError
            An invalid argument was passed.
        ZiplineError
            The server returned the invites in an unexpected format.
        BadRequest
            The server could not process the request.
        Forbidden
            You are not an administrator and cannot use this method.
        """
        if max_uses and not max_uses >= 1:
            raise ValueError("max_uses must be greater than or equal to 1.")

        if expires_at is None:
            expiration = "never"
        elif isinstance(expires_at, datetime.timedelta):
            future_dt = utcnow() + expires_at
            expiration = f"date={to_iso_format(future_dt)}"
        elif isinstance(expires_at, datetime.datetime):
            expiration = f"date={to_iso_format(expires_at)}"

        data: Dict[str, Union[str, int]] = {"expiresAt": expiration}

        if max_uses is not None:
            data["maxUses"] = max_uses

        r = Route("POST", "/api/auth/invites")
        js = await self.http.request(r, json=data)
        return Invite._from_data(js, http=self.http)

    async def delete_invite(self, id: str, /) -> Invite:
        """|coro|

        Deletes an invite with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the invite to delete.

        Returns
        -------
        :class:`~zipline.models.Invite`
            The deleted invite.

        Raises
        ------
        Forbidden
            You are not an administrator and cannot use this method.
        NotFound
            No invite was found with the provided id.
        """
        r = Route("DELETE", f"/api/auth/invites/{id}")
        js = await self.http.request(r)
        return Invite._from_data(js, http=self.http)

    async def get_all_folders(self, *, with_files: bool = True) -> List[Folder]:
        """|coro|

        Returns all folders.

        Parameters
        ----------
        with_files: Optional[:class:`bool`]
            Whether the retrieved folder should contain file information, by default True.

        Returns
        -------
        List[:class:`~zipline.models.Folder`]
            The retrieved Folders.
        """

        params = {}
        if not with_files:
            params["noincl"] = "true"

        r = Route("GET", "/api/user/folders")
        js = await self.http.request(r, params=params)
        return [Folder._from_data(data, http=self.http) for data in js]

    async def create_folder(
        self, name: str, /, files: Optional[List[Union[File, str]]] = None, public: bool = False
    ) -> Folder:
        """|coro|

        Creates a folder.

        Parameters
        ----------
        name: :class:`str`
            The name of the folder to create.
        public: :class:`bool`
            Whether the created folder should be public, by default False.
        files: Optional[List[Union[:class:`~zipline.models.File`, :class:`str`]]]
            Files that should be added to the created folder, if given.

        Returns
        -------
        :class:`~zipline.models.Folder`
            The created folder.

        Raises
        ------
        BadRequest
            The server could not process the request.
        """
        data = {"name": name, "isPublic": public}

        if files:
            file_ids = [file.id if isinstance(file, File) else file for file in files]
            data["files"] = file_ids

        r = Route("POST", "/api/user/folders")
        js = await self.http.request(r, json=data)
        return Folder._from_data(js, http=self.http)

    async def get_folder(self, id: str, /) -> Folder:
        """|coro|

        Gets a folder with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the folder to get.

        Returns
        -------
        :class:`~zipline.models.Folder`
            The requested folder.

        Raises
        ------
        Forbidden
            You do not have access to the folder requested.
        NotFound
            A folder with that id could not be found.
        """
        r = Route("GET", f"/api/user/folders/{id}")
        js = await self.http.request(r)
        return Folder._from_data(js, http=self.http)

    async def get_all_urls(self) -> List[URL]:
        """|coro|

        Retrieves all shortened urls for your user.

        Returns
        -------
        List[:class:`~zipline.models.URL`]
            The requested shortened urls.
        """
        r = Route("GET", "/api/user/urls")
        js = await self.http.request(r)
        return [URL._from_data(data, http=self.http) for data in js]

    async def shorten_url(
        self,
        original_url: str,
        *,
        vanity: Optional[str] = None,
        max_views: Optional[int] = None,
        password: Optional[str] = None,
        enabled: bool = True,
    ) -> URL:
        """|coro|

        Shortens a url.

        Parameters
        ----------
        original_url: :class:`str`
            The url to shorten.
        vanity: Optional[:class:`str`]
            A vanity name to use. None to receive a randomly assigned name.
        max_views: Optional[:class:`int`]
            The number of times the url can be used before being deleted. None for unlimited uses, by default None
        password: Optional[:class:`str`]
            The password required to use the URL, if given.

            .. versionadded:: 0.21.0
        enabled: :class:`bool`
            Whether the url should be enabled for use, by default True.

            .. versionadded:: 0.21.0
        zero_width_space: Optional[:class:`bool`]
            Whether to incude zero width spaces in the returned url, by default False

            .. versionremoved:: 0.21.0
                This is no longer supported by the API.

        Returns
        -------
        :class:`str`
            The shortened url.

        Raises
        ------
        ValueError
            Invalid value for max views passed.
        BadRequest
            The server could not process your request.
        """
        if max_views is not None and max_views < 0:
            raise ValueError("max_views must be greater than or equal to 0")

        headers = {}

        if max_views:
            headers["X-Zipline-Max-Views"] = str(max_views)
        if password:
            headers["X-Zipline-Password"] = password

        data = {"url": original_url, "vanity": vanity, "enabled": enabled}

        r = Route("POST", "/api/user/urls")
        js = await self.http.request(r, headers=headers, json=data)
        return URL._from_data(js, http=self.http)

    async def delete_url(self, id: str, /) -> URL:
        """|coro|

        Delete a url by given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the URL to delete.

        Returns
        -------
        :class:`~zipline.models.URL`
            The most recent information about the deleted url.

        Raises
        ------
        Forbidden
            You do not have access to this url.
        NotFound
            A url with that id could not be found.
        """
        r = Route("DELETE", f"/api/user/urls/{id}")
        js = await self.http.request(r)
        return URL._from_data(js, http=self.http)

    async def get_all_tags(self) -> List[Tag]:
        """|coro|

        Get all tags belonging to the current user.

        Returns
        -------
        List[:class:`~zipline.models.Tag`]
            The requested tags.
        """
        r = Route("GET", "/api/user/tags")
        js = await self.http.request(r)
        return [Tag._from_data(d, http=self.http) for d in js]

    async def get_tag(self, id: str, /) -> Tag:
        """|coro|

        Get a tag with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the tag to retrieve.

        Returns
        -------
        :class:`~zipline.models.Tag`
            The requested tag.

        Raises
        ------
        NotFound
            A tag with that id was not found.
        Forbidden
            You do not have access to this tag.
        """
        r = Route("GET", f"/api/user/tags/{id}")
        js = await self.http.request(r)
        return Tag._from_data(js, http=self.http)

    async def delete_tag(self, id: str, /) -> Tag:
        """|coro|

        Delete a tag with given id.

        Parameters
        ----------
        id: :class:`str`
            The id of the tag to delete.

        Returns
        -------
        :class:`~zipline.models.Tag`
            The deleted tag.

        Raises
        ------
        Forbidden
            You do not have access to this url.
        NotFound
            A url with that id could not be found.
        """
        r = Route("DELETE", f"/api/user/tags/{id}")
        js = await self.http.request(r)
        return Tag._from_data(js, http=self.http)

    # TODO get_user_files or User.get_files is possible by passing an `id` param to /api/user/files with the value of a user id.
    async def get_files(
        self,
        *,
        page: int = 1,
        per_page: int = 10,
        filter: RecentFilesFilter = RecentFilesFilter.all,
        favorite: Optional[bool] = None,
        sort_by: FileSearchSort = FileSearchSort.created_at,
        order: Order = Order.asc,
        search_field: FileSearchField = FileSearchField.file_name,
        search_query: Optional[str] = None,
    ) -> UserFilesResponse:
        """|coro|

        Get files belonging to your user.

        Parameters
        ----------
        page: :class:`int`
            The page of files to get.
        per_page: :class:`int`
            How many files should be returned per page.
        filter: :class:`~zipline.enums.RecentFilesFilter`
            A filter to apply to the files retrieved.
        favorite: Optional[:class:`bool`]
            Whether to search for only favorited or unfavorited files. None for all files.
        sort_by: :class:`~zipline.enums.FileSearchSort`
            How the results should be sorted. Defaults to creation datetime.
        order: :class:`~zipline.enums.Order`
            How the results should be ordered. Defaults to ascending.
        search_field: :class:`~zipline.enums.FileSearchField`
            What file attribute to search by. Defaults to file name.
        search_query: Optional[:class:`str`]
            The query to use in the search.

        Returns
        -------
        :class:`~zipline.models.UserFilesResponse`
            The requested search results.
        """
        params = {"sortBy": sort_by.value, "order": order.value}

        if page:
            params["page"] = str(page)
        if per_page:
            params["perpage"] = str(per_page)
        if filter:
            params["filter"] = filter.value
        if favorite:
            params["favorite"] = "true" if favorite else "false"
        if search_field:
            params["searchField"] = search_field.value
        if search_query:
            params["searchQuery"] = search_query

        r = Route("GET", "/api/user/files")
        js = await self.http.request(r, params=params)
        return UserFilesResponse._from_data(js)

    async def get_recent_files(
        self,
        *,
        amount: int = 10,
        filter: RecentFilesFilter = RecentFilesFilter.all,
    ) -> List[File]:
        """|coro|

        Gets recent files uploaded by you.

        Parameters
        ----------
        amount: Optional[:class:`int`]
            The number of results to return, by default 10.
        filter: :class:`~zipline.enums.RecentFilesFilter`
            What files to get.

            .. versionchanged:: 0.21.0

                This is now controlled by an enum.
        Returns
        -------
        List[:class:`~zipline.models.File`]
            The requested Files.

        Raises
        ------
        ValueError
            An invalid amount was passed.
        """
        if amount < 0:
            raise ValueError("amount must be greater than zero.")

        query_params = {"take": amount, "filter": filter.value}
        r = Route("GET", "/api/user/recent")
        js = await self.http.request(r, params=query_params)
        return [File._from_data(data, http=self.http) for data in js]

    async def upload_file(  # Preliminary done, maybe not
        self,
        payload: FileData,
        *,
        format: NameFormat = NameFormat.uuid,
        compression_percent: int = 0,
        expiry: Optional[datetime.datetime] = None,
        password: Optional[str] = None,
        max_views: Optional[int] = None,
        override_name: Optional[str] = None,
        original_name: Optional[str] = None,
        folder: Optional[Union[Folder, str]] = None,
        override_extension: Optional[str] = None,
    ) -> (
        UploadResponse
    ):  # TODO x-zipline-no-json can return only a url, maybe support this with an override? Also, x-zipline-domain?
        """|coro|

        Upload a file to Zipline.

        Parameters
        ----------
        payload: :class:`~zipline.models.FileData`
            Data regarding the file to upload.
        format: Optional[:class:`~zipline.enums.NameFormat`]
            The format of the name to assign to the uploaded file, by default :attr:`~zipline.enums.NameFormat.uuid`.
        compression_percent: Optional[:class:`int`]
            How compressed should the uploaded file be, by default 0.
        expiry: Optional[:class:`datetime.datetime`]
            When the uploaded file should expire, by default None.
        password: Optional[:class:`str`]
            The password required to view the uploaded File, by default None.
        zero_width_space: Optional[:class:`bool`]
            Whether to include zero width spaces in the name of the uploaded File, by default False.

            .. versionremoved:: 0.21.0
                No longer supported by the API.
        embed: Optional[:class:`bool`]
            Whether to include embed data for the uploaded File, typically used on Discord, by default False.

            .. versionremoved:: 0.21.0
                No longer supported by the API.
        max_views: Optional[:class:`int`]
            The number of times the uploaded File can be viewed before it is deleted, by default None.
        text: Optional[:class:`bool`]
            Whether the File is a text file, by default False.

            .. versionremoved:: 0.21.0
                No longer supported by the API.
        override_name: Optional[:class:`str`]
            A name to give the uploaded file. If provided this will override the server generated name, by default None.
        original_name: Optional[:class:`str`]
            The original_name of the file. None to not preserve this data, by default None.
        folder: Optional[Union[:class:`~zipline.models.Folder`, :class:`str`]]
            The Folder (or it's ID) to place this upload into automatically.

            .. versionadded:: 0.15.0
        override_extension: Optional[:class:`str`]
            The extension to use for this file instead of the original.

            .. versionadded:: 0.21.0

        Returns
        -------
        :class:`~zipline.models.UploadResponse`
            Information about the file uploaded.

        Raises
        ------
        ValueError
            compression_percent was not in 0 <= compression_percent <= 100.
        ValueError
            max_views passed was less than 0.
        ValueError
            The type of the object passed to the folder parameter was incorrect.
        BadRequest
            Server could not process the request.
        ServerError
            The server responded with a 5xx error code.
        """
        if compression_percent < 0 or compression_percent > 100:
            raise ValueError("compression_percent must be between 0 and 100")

        if max_views and max_views <= 0:
            raise ValueError("max_views must be greater than 0")

        headers = {
            "X-Zipline-Format": format.value,
            "X-Zipline-Image-Compression-Percent": str(compression_percent),
        }

        if expiry:
            headers["X-Zipline-Deletes-At"] = f"date={expiry.isoformat()}" if expiry is not None else ""
        if password:
            headers["X-Zipline-Password"] = password if password is not None else ""
        if max_views:
            headers["X-Zipline-Max-Views"] = str(max_views) if max_views is not None else ""
        if override_name:
            headers["X-Zipline-Filename"] = override_name
        if original_name:
            headers["X-Zipline-Original-Name"] = original_name
        if override_extension:
            headers["X-Zipline-File-Extension"] = override_extension
        if folder:
            headers["X-Zipline-Folder"] = folder.id if isinstance(folder, Folder) else folder

        formdata = aiohttp.FormData()
        formdata.add_field("file", payload.data, filename=payload.filename, content_type=payload.mimetype)

        r = Route("POST", "/api/upload")
        js = await self.http.request(r, headers=headers, data=formdata)
        return UploadResponse._from_data(js, http=self.http)

    async def close(self) -> None:
        """|coro|

        Gracefully close the client.
        """
        await self.http.close()

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        await self.close()
